(function() {
  window.PBS = window.PBS || {};
  window.PBS.KIDS = window.PBS.KIDS || {};
  window.PBS.KIDS.SaveSlots = window.PBS.KIDS.SaveSlots || {};

  /**
   * The local storage prefix for fetching keys
   * @var LOCAL_STORAGE_PREFIX
   */
  window.PBS.KIDS.SaveSlots.LOCAL_STORAGE_PREFIX = "pbs-avatar-";

  /**
   * Promised-base GET request implementation
   * @param {string} url The url to fetch
   * @param {string} method GET or POST
   * @param {Object} query An object of query parameters to pass
   * @return {Promise<XMLHttpRequest>} The response
   */
  window.PBS.KIDS.fetch = function(url, method, query) {
    method = method || "GET";
    query = query || {};

    var queryString = Object.keys(query)
      .map(function(key) {
        return encodeURIComponent(key) + "=" + encodeURIComponent(query[key]);
      })
      .join("&");

    return new Promise(function(resolve, reject) {
      var xhr = new XMLHttpRequest();

      xhr.onload = function() {
        if (xhr.status >= 200 && xhr.status < 300) {
          resolve(xhr);
        } else {
          reject(xhr);
        }
      };

      xhr.onerror = function() {
        reject(xhr);
      };

      if (method === "GET") {
        xhr.open("GET", url + "?" + queryString);
        xhr.send();
      } else if (method === "POST") {
        xhr.open("POST", url);
        xhr.setRequestHeader("Content-type", "application/json;");
        xhr.send(JSON.stringify(query));
      } else {
        reject(
          new Error(
            "Unsupported XHR method: " +
              method +
              ". Must be all caps POST or GET."
          )
        );
      }
    });
  };

  /**
   * Loops over a NodeList and executes a callback for each one
   * @param {NodeList} nodeList The nodes to loop over
   * @param {Function} callback The callback to execute
   * @return {void}
   */
  window.PBS.KIDS.forEachNode = function(nodeList, callback) {
    for (var i = 0; i < nodeList.length; i++) {
      callback(nodeList[i], i, nodeList);
    }
  };

  /**
   * Helper method for querying things within this view ala $(selector, el).
   * @param {string} selector A CSS selector to search by
   * @return {Node|null}
   */
  window.PBS.KIDS.SaveSlots.find = function(selector) {
    return PBS.KIDS.SaveSlots.container.querySelector(selector);
  };

  /**
   * Helper method for querying things within this view
   * @param {string} selector A CSS selector
   * @return {NodeList}
   */
  window.PBS.KIDS.SaveSlots.find.all = function(selector) {
    return PBS.KIDS.SaveSlots.container.querySelectorAll(selector);
  };
})();
(function() {
  // formats a byte as a 2-digit hex value
  var toHex = function(byteValue) {
    var rawHex = byteValue.toString(16).toUpperCase();
    if (rawHex.length < 2) {
      rawHex = "0" + rawHex;
    }
    return rawHex;
  };

  /**
   * A UUID v4 generation algorithm, as per https://www.cryptosys.net/pki/uuid-rfc4122.html
   * @return {String} A v4 UUID
   */
  PBS.KIDS.uuid = function() {
    var bytes = new Uint8Array(16);
    crypto.getRandomValues(bytes);

    bytes[6] = 0x40 + (bytes[6] & 0x0f);
    bytes[8] = 0x80 + (bytes[8] & 0x3f);

    var hex = [];
    bytes.forEach(function(byteValue) {
      hex.push(toHex(byteValue));
    });

    return (
      hex[0] +
      hex[1] +
      hex[2] +
      hex[3] +
      "-" +
      hex[4] +
      hex[5] +
      "-" +
      hex[6] +
      hex[7] +
      "-" +
      hex[8] +
      hex[9] +
      "-" +
      hex[10] +
      hex[11] +
      hex[12] +
      hex[13] +
      hex[14] +
      hex[15]
    );
  };
})();
(function() {
  var localStoragePrefix = PBS.KIDS.SaveSlots.LOCAL_STORAGE_PREFIX;

  var findSlot = function(index) {
    return PBS.KIDS.SaveSlots.find('.save-slot[data-index="' + index + '"]');
  };

  var writeSaveSlotDom = function(saveSlot, avatar) {
    saveSlot.className = "save-slot " + avatar.background;
    saveSlot.innerHTML = avatar.html;
    saveSlot.classList.remove("empty");
  };

  var clearSaveSlotDom = function(saveSlot) {
    var path = PBS.KIDS.SaveSlots.config.path + "img/plus.svg";
    saveSlot.innerHTML = '<img alt="+" src="' + path + '" />';
    saveSlot.classList.add("empty");
  };

  /**
   * Saves a given avatar to a save slot, re-rendering it to the DOM
   * @param {Number} slot The 0-indexed slot number
   * @param {String} avatar The HTML for the avatar
   */
  PBS.KIDS.SaveSlots.saveAvatar = function(slot, avatar) {
    localStorage.setItem(localStoragePrefix + slot, JSON.stringify(avatar));
    var saveSlot = findSlot(slot);
    writeSaveSlotDom(saveSlot, avatar);
    PBS.KIDS.SaveSlots.rerenderDeleteButton();
  };

  /**
   * Deletes a save slot, and then re-renders the newly empty slot
   * @param {Number} slot The 0-indexed slot number
   */
  PBS.KIDS.SaveSlots.deleteAvatar = function(slot) {
    localStorage.removeItem(localStoragePrefix + slot);
    var saveSlot = findSlot(slot);
    clearSaveSlotDom(saveSlot);
    PBS.KIDS.SaveSlots.rerenderDeleteButton();
  };

  /**
   * Loads a save slot from local storage and re-renders it
   * @param {Number} slot The 0-indexed slot number
   */
  PBS.KIDS.SaveSlots.reloadAvatar = function(slot) {
    var avatarJson = localStorage.getItem(localStoragePrefix + slot);
    var saveSlot = findSlot(slot);
    if (avatarJson === null) {
      clearSaveSlotDom(saveSlot);
    } else {
      writeSaveSlotDom(saveSlot, JSON.parse(avatarJson));
    }
  };
})();
(function() {
  PBS.KIDS.SaveSlots.getShadowAccount = function(endpoint, apiKey, saveSlot) {
    // TODO: Integrate the saveSlot here. Can't at time of writing, waiting on PRIVO to implement account switching
    var localStorageKey =
      PBS.KIDS.SaveSlots.LOCAL_STORAGE_PREFIX + "shadow-account";

    // if we've already gotten a shadow account, just return the one we have
    if (localStorage.getItem(localStorageKey) !== null) {
      return Promise.resolve(JSON.parse(localStorage.getItem(localStorageKey)));
    }

    // otherwise, ask PRIVO for one
    var url = "https://" + endpoint + "/api/account/shadow?apiKey=" + apiKey;
    var query = {
      roleIdentifier: "pbsFetch_Elem"
    };

    return PBS.KIDS.fetch(url, "POST", query).then(function(xhr) {
      var shadowAccount = JSON.parse(xhr.responseText).entity;
      localStorage.setItem(localStorageKey, JSON.stringify(shadowAccount));
      return shadowAccount;
    });
  };
})();
(function() {
  PBS.KIDS.SaveSlots.enterDeleteMode = function(e) {
    PBS.KIDS.SaveSlots.container.classList.add("delete-mode");

    // if this enter delete mode was caused by a event, don't propagate because it'll trigger a exitDeleteMode automatically
    if (e && e.stopPropagation instanceof Function) {
      e.stopPropagation();
    }
  };

  PBS.KIDS.SaveSlots.exitDeleteMode = function() {
    PBS.KIDS.SaveSlots.container.classList.remove("delete-mode");
    PBS.KIDS.SaveSlots.find(".modal-wrapper").classList.add("hidden");
  };

  PBS.KIDS.SaveSlots.deleteCurrentlySelectedSlot = function() {
    var index = PBS.KIDS.SaveSlots.find(".modal").dataset.saveSlotIndex;
    PBS.KIDS.SaveSlots.deleteAvatar(index);
    PBS.KIDS.SaveSlots.exitDeleteMode();
  };

  PBS.KIDS.SaveSlots.rerenderDeleteButton = function() {
    var deleteButton = PBS.KIDS.SaveSlots.find(".delete");
    deleteButton.classList.remove("hidden");
    if (PBS.KIDS.SaveSlots.find(".save-slot:not(.empty)") === null) {
      deleteButton.classList.add("hidden");
    }
  };
})();
(function() {
  PBS.KIDS.SaveSlots.editSaveSlot = function(index) {
    // mark the avatar button with the current save slot that's being updated
    PBS.KIDS.SaveSlots.find(".ok-button").dataset.index = index;

    // reset the avatar editor
    PBS.KIDS.SaveSlots.resetAvatarEditor();

    // now, move to that screen
    PBS.KIDS.SaveSlots.find(".save-slots").classList.add("hidden");
    PBS.KIDS.SaveSlots.find(".avatar").classList.remove("hidden");
  };

  PBS.KIDS.SaveSlots.chooseSaveSlot = function(index) {
    var element = PBS.KIDS.SaveSlots.find(
      '.save-slot[data-index="' + index + '"]'
    );

    var isEmptySlot = element.classList.contains("empty");
    var inDeleteMode = PBS.KIDS.SaveSlots.container.classList.contains(
      "delete-mode"
    );

    if (isEmptySlot && !inDeleteMode) {
      PBS.KIDS.SaveSlots.editSaveSlot(index);
    } else if (!isEmptySlot && inDeleteMode) {
      // show the modal to delete the slot
      var modal = (PBS.KIDS.SaveSlots.find(
        ".modal"
      ).dataset.saveSlotIndex = index);
      PBS.KIDS.SaveSlots.find(".modal-wrapper").classList.remove("hidden");
    } else if (!isEmptySlot && !inDeleteMode) {
      PBS.KIDS.SaveSlots.avatar = JSON.parse(
        localStorage.getItem(PBS.KIDS.SaveSlots.LOCAL_STORAGE_PREFIX + index)
      );
      PBS.KIDS.SaveSlots.avatar.index = index;

      PBS.KIDS.SaveSlots.getShadowAccount(
        PBS.KIDS.SaveSlots.config.privoEndpoint,
        PBS.KIDS.SaveSlots.config.privoApiKey,
        index
      )
        .catch(function(error) {
          console.log("PRIVO Shadow account creation failed", error);
          return null;
        })
        .then(function(shadowAccount) {
          PBS.KIDS.SaveSlots.shadowAccount = shadowAccount;
          PBS.KIDS.SaveSlots._listeners.onSelectAvatar(
            PBS.KIDS.SaveSlots.avatar,
            shadowAccount
          );
        });
    } else {
      PBS.KIDS.SaveSlots.exitDeleteMode();
    }
  };
})();
(function() {
  var find = PBS.KIDS.SaveSlots.find;

  var colors = {
    background: [
      "magenta",
      "green",
      "seafoam",
      "blue",
      "yellow",
      "pink",
      "orange"
    ],
    face: [
      "light_green",
      "dark_green",
      "light_blue",
      "dark_blue",
      "yellow",
      "purple",
      "orange",
      "dark_orange",
      "light_seafoam",
      "seafoam",
      "light_pink",
      "dark_pink"
    ],
    eye: [
      "black",
      "blue",
      "dark-orange",
      "pink",
      "green",
      "orange",
      "purple",
      "seafoam"
    ],
    hair: ["black", "blue", "green", "red", "yellow", "purple"]
  };

  /**
   * Removes the current color class from the element, then replaces it with the next one in the list, wrapping around
   * if necessary
   *
   * @param {Node} element The element to swap the class on
   * @param {Array<string>} colors The array of colors to look through for the next color
   * @return void
   */
  var moveToNextColor = function(element, colors) {
    var index;
    for (var i = 0; i < element.classList.length; i++) {
      index = colors.indexOf(element.classList[i]);
      if (index > -1) {
        break;
      }
    }

    // Couldn't find a color, so this is a no-op
    if (index === -1) {
      console.warn(
        "Attempted to find a color on",
        element,
        "from list",
        colors,
        "but could not find anything"
      );
      return;
    }

    // now, swap out the colors, since we found an index
    var nextIndex = (index + 1) % colors.length;
    element.classList.remove(colors[index]);
    element.classList.add(colors[nextIndex]);
  };

  /**
   * Gets a 0-padded index between 1 and max, inclusive
   * @param {Number} max The upper bound
   * @return {String}
   */
  var randomPaddedIndex = function(max) {
    var index = Math.floor(Math.random() * max) + 1;
    if (index < 10) {
      return "0" + index;
    }

    return "" + index;
  };

  PBS.KIDS.SaveSlots.nextHair = function() {
    var hairDo = "img/hair/hair-" + randomPaddedIndex(28) + ".svg";
    var randomColor =
      colors.hair[Math.floor(Math.random() * colors.hair.length)];
    PBS.KIDS.SaveSlots.renderAvatarPart("hair", hairDo, randomColor);
  };

  PBS.KIDS.SaveSlots.nextFace = function() {
    var randomColor =
      colors.face[Math.floor(Math.random() * colors.face.length)];
    PBS.KIDS.SaveSlots.renderAvatarPart("face", "img/head.svg", randomColor);
  };

  PBS.KIDS.SaveSlots.nextEyes = function() {
    var eyeImage = "img/eyes/eyes-" + randomPaddedIndex(12) + ".svg";
    var randomColor = colors.eye[Math.floor(Math.random() * colors.eye.length)];
    PBS.KIDS.SaveSlots.renderAvatarPart("eyes", eyeImage, randomColor);
  };

  PBS.KIDS.SaveSlots.nextBackground = function() {
    var randomColor =
      colors.background[Math.floor(Math.random() * colors.background.length)];
    var avatarPreview = find(".avatar-preview .background-container");
    avatarPreview.className = "background-container " + randomColor;
  };

  PBS.KIDS.SaveSlots.renderAvatarPart = function(containerType, asset, color) {
    var containerClass = containerType + "-container";
    var assetClass = asset.replace(/.*\//, "").replace(/\.svg$/, "");
    var container = find(".avatar-preview ." + containerClass);
    container.innerHTML = PBS.KIDS.SaveSlots.assets[asset];
    container.className = containerClass + " " + color + " " + assetClass;
  };

  PBS.KIDS.SaveSlots.resetAvatarEditor = function() {
    PBS.KIDS.SaveSlots.nextHair();
    PBS.KIDS.SaveSlots.nextFace();
    PBS.KIDS.SaveSlots.nextEyes();
    PBS.KIDS.SaveSlots.nextBackground();
  };

  PBS.KIDS.SaveSlots.serializeAvatarEditor = function() {
    // find the first background container class that's not "background-container", which should be a color
    var backgroundContainer = find(".avatar-preview .background-container");
    var backgroundColor = "";
    for (var i = 0; i < backgroundContainer.classList.length; i++) {
      if (backgroundContainer.classList[i] !== "background-container") {
        backgroundColor = backgroundContainer.classList[i];
        break;
      }
    }

    return {
      html: find(".avatar-preview").innerHTML,
      background: backgroundColor,
      id: PBS.KIDS.uuid()
    };
  };
})();
(function() {
  var preloadItems = [
    "avatar-save-slot.html",
    "img/arrow.svg",
    "img/background.svg",
    "img/checkmark.svg",
    "img/ex.svg",
    "img/eyes.svg",
    "img/hair.svg",
    "img/head.svg",
    "img/plus.svg",
    "img/trash.svg"
  ];

  for (var i = 1; i <= 12; i++) {
    var path = "img/eyes/eyes-";
    if (i < 10) {
      path += "0";
    }
    path += i + ".svg";
    preloadItems.push(path);
  }

  for (var i = 1; i <= 28; i++) {
    var path = "img/hair/hair-";
    if (i < 10) {
      path += "0";
    }
    path += i + ".svg";
    preloadItems.push(path);
  }

  var find = PBS.KIDS.SaveSlots.find;

  var onClickAvatarBack = function(evt) {
    find(".ok-button").dataset.index = "";
    find(".save-slots").classList.remove("hidden");
    find(".avatar").classList.add("hidden");
  };

  var onClickAvatarOk = function(evt) {
    var serializedAvatar = PBS.KIDS.SaveSlots.serializeAvatarEditor();
    var index = find(".ok-button").dataset.index;

    PBS.KIDS.SaveSlots.saveAvatar(index, serializedAvatar);

    find(".ok-button").dataset.index = "";
    find(".save-slots").classList.remove("hidden");
    find(".avatar").classList.add("hidden");
  };

  /**
   * The main container for this view, should be an element with id=avatar-save-slot
   * @property {Node}
   */
  PBS.KIDS.SaveSlots.container = null;

  /**
   * Builds the view and sets up event listeners, but does not show it
   * @param {Object} config A object with configuration in it
   * @param {String} config.path The path to load assets from
   * @param {String} config.privoEndpoint The endpoint for authenticating with privo
   * @param {String} config.privoApiKey The API key for authenticating with privo
   * @param {Object} config.listeners An object containing listener functions for events this module raises
   * @param {Function} config.listeners.onBack A callback for when the user backs out of the view
   * @param {Function} config.listeners.onSelectAvatar A callback for when the user selects an avatar
   */
  PBS.KIDS.SaveSlots.init = function(config) {
    config = config || {};
    config.listeners = config.listeners || {};
    config.path = config.path || "dist/";
    config.privoEndpoint = config.privoEndpoint || "privohub-staging.privo.com";
    config.privoApiKey = config.privoApiKey || "invalidkey";
    if (config.path[config.path.length - 1] !== "/") {
      config.path += "/";
    }

    PBS.KIDS.SaveSlots.config = config;

    PBS.KIDS.SaveSlots.container = document.querySelector("#avatar-save-slot");
    PBS.KIDS.SaveSlots.container.classList.add("hidden");

    PBS.KIDS.SaveSlots._listeners = config.listeners;
    PBS.KIDS.SaveSlots._listeners.onBack =
      config.listeners.onBack || function() {};
    PBS.KIDS.SaveSlots._listeners.onSelectAvatar =
      config.listeners.onSelectAvatar || function() {};

    PBS.KIDS.SaveSlots.assets = {};

    var loads = preloadItems.map(function(item) {
      return PBS.KIDS.fetch(config.path + item);
    });

    return Promise.all(loads)
      .then(function(xhrs) {
        for (var i = 0; i < xhrs.length; i++) {
          var assetName = preloadItems[i];
          var responseText = xhrs[i].responseText;
          PBS.KIDS.SaveSlots.assets[assetName] = responseText;
        }

        return PBS.KIDS.SaveSlots.assets;
      })
      .then(function() {
        PBS.KIDS.SaveSlots.container.innerHTML =
          PBS.KIDS.SaveSlots.assets["avatar-save-slot.html"];

        // event listeners
        var saveSlots = find.all(".save-slot");
        PBS.KIDS.forEachNode(saveSlots, function(node) {
          PBS.KIDS.SaveSlots.reloadAvatar(node.dataset.index);

          node.addEventListener("click", function(e) {
            PBS.KIDS.SaveSlots.chooseSaveSlot(node.dataset.index);

            // don't propagate, because that will automatically cancel delete
            // mode. chooseSaveSlot has specific logic for handling that case
            // instead
            e.stopPropagation();
          });
        });
        PBS.KIDS.SaveSlots.container.addEventListener(
          "click",
          PBS.KIDS.SaveSlots.exitDeleteMode
        );
        find(".delete").addEventListener(
          "click",
          PBS.KIDS.SaveSlots.enterDeleteMode
        );
        find(".modal").addEventListener("click", function(e) {
          // prevent user clicks for cancelling delete mode
          e.stopPropagation();
        });
        find(".modal .yes").addEventListener(
          "click",
          PBS.KIDS.SaveSlots.deleteCurrentlySelectedSlot
        );
        find(".modal .no").addEventListener(
          "click",
          PBS.KIDS.SaveSlots.exitDeleteMode
        );
        find(".avatar .back").addEventListener("click", onClickAvatarBack);
        find(".avatar .ok-button").addEventListener("click", onClickAvatarOk);
        find(".save-slots .back").addEventListener(
          "click",
          PBS.KIDS.SaveSlots._listeners.onBack
        );
        find(".avatar .avatar-button.hair").addEventListener(
          "click",
          PBS.KIDS.SaveSlots.nextHair
        );
        find(".avatar .avatar-button.face").addEventListener(
          "click",
          PBS.KIDS.SaveSlots.nextFace
        );
        find(".avatar .avatar-button.eye").addEventListener(
          "click",
          PBS.KIDS.SaveSlots.nextEyes
        );
        find(".avatar .avatar-button.background").addEventListener(
          "click",
          PBS.KIDS.SaveSlots.nextBackground
        );

        // add some icons to everything
        find(".save-slots .back img").src = config.path + "img/arrow.svg";
        find(".save-slots .delete img").src = config.path + "img/trash.svg";
        find(".modal .yes img").src = config.path + "img/checkmark.svg";
        find(".modal .no img").src = config.path + "img/ex.svg";
        find(".avatar .back img").src = config.path + "img/arrow.svg";
        find(".avatar .avatar-button.eye img").src =
          config.path + "img/eyes.svg";
        find(".avatar .avatar-button.hair img").src =
          config.path + "img/hair.svg";
        find(".avatar .avatar-button.face img").src =
          config.path + "img/head.svg";
        find(".avatar .avatar-button.background img").src =
          config.path + "img/background.svg";
        find(".avatar .ok-button img").src = config.path + "img/checkmark.svg";

        // prevent any images from being draggable
        PBS.KIDS.forEachNode(find.all("img"), function(image) {
          image.draggable = false;
        });

        PBS.KIDS.SaveSlots.rerenderDeleteButton();
      });
  };

  /**
   * Shows the view
   */
  PBS.KIDS.SaveSlots.show = function() {
    PBS.KIDS.SaveSlots.container.classList.remove("hidden");
    find(".save-slots").classList.remove("hidden");
    find(".avatar").classList.add("hidden"); // always want this hidden at first
  };

  /**
   * Hides the view
   */
  PBS.KIDS.SaveSlots.hide = function() {
    PBS.KIDS.SaveSlots.container.classList.add("hidden");
    find(".save-slots").classList.add("hidden");
    find(".avatar").classList.add("hidden");
  };

  /**
   * Removes everything from the container
   */
  PBS.KIDS.SaveSlots.destroy = function() {
    // delete everything, which should remove any listeners
    PBS.KIDS.SaveSlots.container.innerHTML = "";
    PBS.KIDS.SaveSlots.container = null;
  };

  /**
   * The currently selected avatar, available even when this view is not showing
   */
  PBS.KIDS.SaveSlots.avatar = null;
})();
